#ifndef __DS485_SOCKET_SERVER_IMPL__
#define __DS485_SOCKET_SERVER_IMPL__

#if HAVE_VISIBILITY
#define NETLIB_EXPORTED(sym) typeof(sym) sym __attribute__((__visibility__("default")));
#else
#define NETLIB_EXPORTED(sym)
#endif

/* only 4 bits for transaction id */
#define MAX_CLIENTS     (1 << 4)

/**
 * How long threads might sleep until checking
 * server->active again
 * will delay shutdown by this many seconds
 */
#define MAX_THREAD_SLEEP           1 /* sec */

/**
 * When to consider a receiver thread to be stuck
 */
#define CLIENT_INACTIVE_THRESHOLD   30 /* sec */

/**
 * Client session
 */
typedef struct _dS485dClient {
    struct _dS485dClient*   next;
    struct dS485_server     *server;

    int fd;                   /* client socket descriptor */
    int rxLen;
    uint8_t  *rxBuffer;

    int transactionId;        /* serves like dsuid of the client, same for all requests of this client */
    pthread_t tid;            /* thread id */

    uint8_t apirev;           /* client API revision */
    uint8_t rxfilter;         /* reception packet filter */

    /*
     * Only lock the socket itself, not the whole structure, tbd
     * http://lwn.net/Articles/180387/
     */
    pthread_mutex_t lock;
    char peer_name[25];

    time_t ts_detach;
    time_t ts_activity;
    int last_command;
} dS485dClient;

/* CAUTION: requires cl->lock to be held */
void _ds485n_client_detach(dS485dClient *cl);

/**
 * Server
 */
struct dS485_server
{
    dS485dClient* clSessions;
    int transid_pool[MAX_CLIENTS]; /* transaction id pool */
    int client_count;
    pthread_rwlock_t clientListRWLock;

    uint64_t rxPacketNum;
    uint64_t txPacketNum;
    bool active;
    bool upperPath;

    /* server socket file descriptor */
    int sfd;
    struct dS485_server_callbacks cb;

    pthread_t poll_tid;            /* thread id */
};


/**
 * header of ds485n_packet_t and ds485_container differs
 * - ds485n_packet_t contains dummy field to ensure alignment
 * - ds485n_packet_t has 16bit length ds485_container only 8bit
 * hence we can not declare a common header struct and
 * cast the container/pointer to that type
 *
 * either we clutter the code with (type == v_container) tests
 * everywhere or we cache the fields and copy them back before
 * before forwarding the containers
 * CAUTION: always call clPacketUnionSync after doing changes
 */
typedef struct {
    dsuid_t src;
    dsuid_t dst;
    unsigned char transactionId;
    uint32_t containerType;

    enum {
        v_container,
        v_packet
    } type;

    union {
        const ds485_container_t *container;
        const ds485n_packet_t *packet;
    } u;
} packet_union_t;

void ds485n_pu_from_short(packet_union_t *pkt, const ds485_container_t *cont);
void ds485n_pu_from_long(packet_union_t *dst, const ds485n_packet_t *src);

/* CAUTION: functions prefixed with '_' require holding a lock */

/**
 * TCP is stream oriented, packetize with tag-length-value
 * _ds485n_pu_write/_ds485n_tlv_write call this function internally
 */
int _ds485n_tlv_write(dS485dClient *cl, uint8_t cmd, const void *data, int dlen);

/**
 * requests from upper dsm-api layer, not really ds485-netlib function
 * could be called container write
 */
int _ds485n_dsmapi_write(dS485dClient *cl, uint8_t cmd,
                            const ds485_container_t *cont);
/**
 * routable data packets
 */
int _ds485n_pu_write(dS485dClient *cl, const packet_union_t *packet);

/**
 * write modified src/transactionId back to either short or long container
 * see description in packet_untiont declaration
 */
void ds485n_pu_sync(packet_union_t *packet);

/**
 * @ret true if size does not exceed upper bound
 */
bool ds485n_pu_is_oversize(const packet_union_t *packet);

/**
 * Forward packet to server and all peers incl. ourselves if promiscous
 */
int ds485n_pu_request(struct dS485_server *server, packet_union_t *packet,
                      dS485dClient *sender);

/**
 * parse tag-length-value chunk of data
 */
int ds485_netlib_packet_parser(dS485dClient *cl, uint8_t* data);

/**
 * Create v1 bus change message
 */
void ds485n_create_bus_change(ds485_container_t *cont, dsuid_t *dsuid,
                              int flags);

/**
 * Create v1 bus state message
 */
void ds485n_create_bus_state(ds485_container_t *cont, bus_state_t state);

#endif
