#include "ds485-socket-server.h"

#include <assert.h>
#include <arpa/inet.h>
#include <stdbool.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <pthread.h>

#include <digitalSTROM/ds485.h>
#include <digitalSTROM/ds485-core.h>
#include "ds485-socket-interface.h"

#include "ds485-socket-server-impl.h"

#include "base/log.h"
#include "base/util.h"

void ds485n_pu_from_short(packet_union_t *pkt, const ds485_container_t *cont)
{
    memcpy(&pkt->src, &cont->sourceId, sizeof(pkt->src));
    memcpy(&pkt->dst, &cont->destinationId, sizeof(pkt->dst));
    pkt->transactionId = cont->transactionId;
    pkt->type = v_container;
    pkt->containerType = cont->containerType;
    pkt->u.container = cont;
}

void ds485n_pu_from_long(packet_union_t *pkt, const ds485n_packet_t *lcont)
{
    memcpy(&pkt->src, &lcont->sourceId, sizeof(pkt->src));
    memcpy(&pkt->dst, &lcont->destinationId, sizeof(pkt->dst));
    pkt->transactionId = lcont->transactionId;
    pkt->type = v_packet;
    pkt->containerType = lcont->containerType;
    // lcont is up to 4kByte, careful with copy!
    pkt->u.packet = lcont;
}

void ds485n_pu_sync(packet_union_t *packet)
{
    if (packet->type == v_container) {
        // cast away const
        ds485_container_t *tmp = (ds485_container_t *)packet->u.container;
        memcpy(&tmp->sourceId, &packet->src, sizeof(packet->src));
        memcpy(&tmp->destinationId, &packet->dst, sizeof(packet->dst));
        tmp->transactionId = packet->transactionId;
        tmp->containerType = packet->containerType;
    } else {
        // cast away const
        ds485n_packet_t *tmp = (ds485n_packet_t *)packet->u.packet;
        memcpy(&tmp->sourceId, &packet->src, sizeof(packet->src));
        memcpy(&tmp->destinationId, &packet->dst, sizeof(packet->dst));
        tmp->transactionId = packet->transactionId;
        tmp->containerType = packet->containerType;
    }
}

bool ds485n_pu_is_oversize(const packet_union_t *packet) {
    return (packet->type == v_container) ?
        packet->u.container->length > DS485_MAX_PAYLOAD_SIZE :
        packet->u.packet->length > DS485N_PACKET_MAX_PAYLOAD_SIZE;
}

static int ds485n_quick_reply_dsmapi(dS485dClient* cl, uint8_t cmd,
                                     const ds485_container_t *container)
{
    int ret;
    pthread_mutex_lock(&cl->lock);
    ret = _ds485n_dsmapi_write(cl, cmd, container);
    pthread_mutex_unlock(&cl->lock);
    return ret;
}

static int ds485n_quick_reply(dS485dClient* cl, uint8_t cmd,
                              const void *data, int dlen)
{
    int ret;
    pthread_mutex_lock(&cl->lock);
    ret = _ds485n_tlv_write(cl, cmd, data, dlen);
    pthread_mutex_unlock(&cl->lock);
    return ret;
}

/* CAUTION: client->lock must be held */
int _ds485n_pu_write(dS485dClient* cl, const packet_union_t *packet)
{
    // we are the server we send only RX packets
    return (packet->type == v_container) ?
        _ds485n_tlv_write(cl, DS485D_CMD_RX_PACKET, packet->u.container,
                          DS485_CONTAINER_SIZE + packet->u.container->length) :
        _ds485n_tlv_write(cl, DS485N_CMD_LONG_RX_PACKET, packet->u.packet,
                          DS485N_HDR_LEN + packet->u.packet->length);
}

void ds485n_create_bus_change(ds485_container_t *cont, dsuid_t *dsuid,
                              int flags)
{
    memset(cont, 0, sizeof(*cont)); /* valgrind */
    memcpy(&cont->sourceId, dsuid, sizeof(*dsuid));
    cont->data[0] = flags;
    cont->length = 1;
}

void ds485n_create_bus_state(ds485_container_t *cont, bus_state_t state)
{
    memset(cont, 0, sizeof(*cont)); /* valgrind */
    cont->data[0] = state;
    cont->length = 1;
}

int ds485_netlib_packet_parser(dS485dClient *cl, uint8_t* data)
{
    struct dS485_server *server = cl->server;
    dS485dCommand* cmd;
    uint8_t value;
    int rval = 0;

    cmd = (dS485dCommand *) data;

    cl->last_command = cmd->command;

    /* TODO, alignment errors when accessing container.containerFlags */

    report2(LOG_DEBUG, cl->peer_name,
            "command: 0x%02x, len 0x%02x, data length %d\n", cmd->command,
           cmd->length, cl->rxLen);

    switch (cmd->command) {

    case DS485D_CMD_API_REVISION:
        cl->apirev = cmd->data[0];
        value = DS485D_API_REVISION;
        ds485n_quick_reply(cl, DS485D_CMD_API_REVISION, &value, 1);
        break;

    case DS485D_CMD_PING:
        report2(LOG_DEBUG, cl->peer_name, "ping received, sending pong\n");
        ds485n_quick_reply(cl, DS485D_CMD_PING, NULL, 0);
        break;

    case DS485D_CMD_TX_PACKET: {
        dS485CmdApidata *apiData = (dS485CmdApidata *) cmd;
        ds485_container_t *container = &apiData->apiData; /* alignment */
        uint16_t pkt_len = apiData->length - 3;
        packet_union_t data;

        if (pkt_len < DS485_CONTAINER_SIZE + container->length) {
            report2(LOG_WARNING, cl->peer_name,
                    "wrapped container size mismatch: container: %d wrapper: %d\n",
                    container->length + DS485_CONTAINER_SIZE, cmd->length);
            return -ENODATA; /* TODO kick client */
        }

        if (container->length > DS485_MAX_PAYLOAD_SIZE) {
            CONTAINER_STR_BUF(buf);
            sprintf_container(buf, container);
            report2(LOG_WARNING, cl->peer_name,
                    "not forwarding oversize container: %s\n", buf);
            return -ENOBUFS;
        }

        ds485n_pu_from_short(&data, container);
        rval = ds485n_pu_request(cl->server, &data, cl);
        break;
    }

    case DS485N_CMD_LONG_TX_PACKET: {
        dS485CmdLongApidata *apiData = (dS485CmdLongApidata *) cmd;
        ds485n_packet_t *packet = &apiData->apiData; /* alignment */
        uint16_t pkt_len = apiData->length - 3;
        packet_union_t data;

        if (pkt_len != DS485N_HDR_LEN + packet->length) {
            report2(LOG_WARNING, cl->peer_name,
                    "wrapped packet size mismatch: packet: %d wrapper %d\n",
                    packet->length + DS485N_HDR_LEN, cmd->length);
            return -ENODATA; /* TODO kick client */
        }

        if (packet->length > DS485N_PACKET_MAX_PAYLOAD_SIZE) {
            CONTAINER_STR_BUF(buf);
            sprintf_netlib_packet(buf, packet);
            report2(LOG_WARNING, cl->peer_name,
                    "not forwarding oversize long packet: %s\n", buf);
            return -ENOBUFS;
        }

        ds485n_pu_from_long(&data, packet);
        rval = ds485n_pu_request(cl->server, &data, cl);
        break;
    }

    case DS485D_CMD_RX_PACKET:
        report2(LOG_ERR, cl->peer_name,
                "received invalid client command DS485D_CMD_RX_PACKET\n");
        break;

    case DS485N_CMD_LONG_RX_PACKET:
        report2(LOG_ERR, cl->peer_name,
                "received invalid client command DS485N_CMD_LONG_RX_PACKET\n");
        break;

    case DS485D_CMD_PACKETFILTER:
        rval = server->cb.set_packetfilter(server, cmd->data[0]);
        if (rval < 0) {
            report2(LOG_WARNING, cl->peer_name,
                    "ds485_server_set_packetfilter failed %d\n", rval);
            /* TODO forward error */
            break;
        }
        cl->rxfilter = cmd->data[0];
        report2(LOG_DEBUG, cl->peer_name, "set rxfilter: 0x%02x\n",
                cl->rxfilter);
        break;

    case DS485D_CMD_QUERY_DEVICES: {
        DSUID_STR_BUF(dsuid_str);
        int length, n, i;
        dsuid_t *devices;

        length = cmd->data[0];
        devices = (dsuid_t *)malloc(length * sizeof(dsuid_t));
        if (!devices) {
            report2(LOG_WARNING, cl->peer_name,
                    "DS485D_CMD_QUERY_DEVICES malloc failed\n");
            /* TODO return error code */
            break;
        }

        n = server->cb.get_bus_members(server, devices, length);

        report2(LOG_INFO, cl->peer_name, "device list of %d devices\n", n);
        for (i = 0; i < n; i++) {
            sprintf_dsuid(dsuid_str, &devices[i]);
            report2(LOG_INFO, cl->peer_name, "     [%d] %s\n", i, dsuid_str);
        }
        /* TODO n < 0, return error code */
        if (n == 0) {
            ds485n_quick_reply(cl, DS485D_CMD_QUERY_DEVICES, NULL, 0);
        } else {
            ds485n_quick_reply(cl, DS485D_CMD_QUERY_DEVICES, devices,
                               n * sizeof(dsuid_t));
        }
        free(devices);
        break;
    }

    case DS485D_CMD_QUERY_DSUID: {
        dsuid_t dsuid = server->cb.get_dsuid(server);
        ds485n_quick_reply(cl, DS485D_CMD_QUERY_DSUID, &dsuid, sizeof(dsuid_t));
        /* TODO error handling */
        break;
    }

    case DS485D_CMD_QUERY_ADDRESS:
        /* TODO: send reply */
        break;

    case DS485D_CMD_QUERY_BUSSTATE:
        cmd->data[0] = server->cb.get_bus_state(server);
        ds485n_quick_reply(cl, DS485D_CMD_QUERY_BUSSTATE, cmd->data, 1);
        /* TODO error handling */
        break;

    case DS485D_CMD_BUS_STATE: {
        ds485_container_t* packet = (ds485_container_t *) &cmd->data[0];
        packet->data[0] = server->cb.get_bus_state(server);
        packet->length = 1;
        ds485n_quick_reply_dsmapi(cl, DS485D_CMD_BUS_STATE, packet);
        /* TODO error handling */
        break;
    }

    case DS485D_CMD_RESTART_BUS:
    {
        ds485_container_t* packet = (ds485_container_t *) &cmd->data[0];
        packet->data[0] = server->cb.restart_bus(server);
        packet->length = 1;
        ds485n_quick_reply_dsmapi(cl, DS485D_CMD_RESTART_BUS, packet);
        /* TODO error handling */
        break;
    }

    case DS485D_CMD_DEVICE_PACKET:
    case DS485N_CMD_LONG_DEVICE_PACKET:
    case DS485D_CMD_RAWMODE:
        report2(LOG_WARNING, cl->peer_name, "deprecated cmd:0x%02x\n",
                cmd->command);
        break;

    default:
        report2(LOG_WARNING, cl->peer_name, "unknown cmd:0x%02x\n",
                cmd->command);
        break;
    }

    return rval;
}
